/*
 * Copyright 2010 Joachim Ansorg, mail@ansorg-it.com
 * File: EvaluateStaticArithExprInspection.java, Class: EvaluateStaticArithExprInspection
 * Last modified: 2010-03-24
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.ansorgit.plugins.bash.editor.inspections.inspections;

import com.ansorgit.plugins.bash.editor.inspections.quickfix.EvaluateArithExprQuickfix;
import com.ansorgit.plugins.bash.lang.psi.BashVisitor;
import com.ansorgit.plugins.bash.lang.psi.api.arithmetic.ArithmeticExpression;
import com.intellij.codeHighlighting.HighlightDisplayLevel;
import com.intellij.codeInspection.ProblemsHolder;
import com.intellij.psi.PsiElementVisitor;
import org.intellij.lang.annotations.Pattern;
import org.jetbrains.annotations.Nls;
import org.jetbrains.annotations.NotNull;

import java.util.List;

/**
 * Evaluate a static arithmetic expression. Offers a quickfix to insert the replacement value.
 * <p/>
 * User: jansorg
 * Date: Nov 15, 2009
 * Time: 12:48:24 AM
 */
public class EvaluateStaticArithExprInspection extends AbstractBashInspection {
    @Pattern("[a-zA-Z_0-9.]+")
    @NotNull
    @Override
    public String getID() {
        return "EvaluateArithmeticExpression";
    }

    @NotNull
    @Override
    public String getShortName() {
        return "Evaluate arithmetic expression";
    }

    @Nls
    @NotNull
    @Override
    public String getDisplayName() {
        return "Evaluate an arithmetic expression";
    }

    @Override
    public String getStaticDescription() {
        return "Replaces a static arithmetic expression with the result. For example the expression 1 + 3 * 4 would be replaced with the evaluated result of 13.";
    }

    @NotNull
    @Override
    public HighlightDisplayLevel getDefaultLevel() {
        return HighlightDisplayLevel.WARNING;
    }

    @NotNull
    @Override
    public PsiElementVisitor buildVisitor(@NotNull final ProblemsHolder holder, final boolean isOnTheFly) {
        return new BashVisitor() {
            @Override
            public void visitArithmeticExpression(ArithmeticExpression expression) {
                List<ArithmeticExpression> subexpressions = expression.subexpressions();
                if (isOnTheFly && subexpressions.size() > 1 && expression.isStatic()) {
                    ArithmeticExpression parent = expression.findParentExpression();
                    if (parent == null || !parent.isStatic()) {
                        try {
                            String template = "Replace with the evaluated result '" + expression.computeNumericValue() + "'";
                            holder.registerProblem(expression, template, new EvaluateArithExprQuickfix(expression));
                        } catch (Exception e) {
                            //fixme
                        }
                    }
                }
            }
        };
    }
}